"""
This module contains condition classes for supervised learning tasks.
"""

import torch
from torch_geometric.data import Data
from ..label_tensor import LabelTensor
from ..graph import Graph
from .condition_interface import ConditionInterface


class InputTargetCondition(ConditionInterface):
    """
    Condition defined by input and target data. This condition can be used in
    both supervised learning and Physics-informed problems. Based on the type of
    the input and target, different condition implementations are available:

    - :class:`TensorInputTensorTargetCondition`: For :class:`torch.Tensor` or \
        :class:`~pina.label_tensor.LabelTensor` input and target data.
    - :class:`TensorInputGraphTargetCondition`: For :class:`torch.Tensor` or \
        :class:`~pina.label_tensor.LabelTensor` input and \
        :class:`~pina.graph.Graph` or :class:`torch_geometric.data.Data` \
        target data.
    - :class:`GraphInputTensorTargetCondition`: For :class:`~pina.graph.Graph` \
        or :class:`~torch_geometric.data.Data` input and :class:`torch.Tensor` \
        or :class:`~pina.label_tensor.LabelTensor` target data.
    - :class:`GraphInputGraphTargetCondition`: For :class:`~pina.graph.Graph` \
        or :class:`~torch_geometric.data.Data` input and target data.
    """

    __slots__ = ["input", "target"]
    _avail_input_cls = (torch.Tensor, LabelTensor, Data, Graph, list, tuple)
    _avail_output_cls = (torch.Tensor, LabelTensor, Data, Graph, list, tuple)

    def __new__(cls, input, target):
        """
        Instantiate the appropriate subclass of InputTargetCondition based on
        the types of input and target data.

        :param input: Input data for the condition.
        :type input: torch.Tensor | LabelTensor | Graph | Data | list[Graph] |
            list[Data] | tuple[Graph] | tuple[Data]
        :param target: Target data for the condition.
        :type target: torch.Tensor | LabelTensor | Graph | Data | list[Graph] |
            list[Data] | tuple[Graph] | tuple[Data]
        :return: Subclass of InputTargetCondition
        :rtype: pina.condition.input_target_condition.
            TensorInputTensorTargetCondition |
            pina.condition.input_target_condition.
            TensorInputGraphTargetCondition |
            pina.condition.input_target_condition.
            GraphInputTensorTargetCondition |
            pina.condition.input_target_condition.GraphInputGraphTargetCondition

        :raises ValueError: If ``input`` and/or ``target`` are not of type
            :class:`torch.Tensor`, :class:`~pina.label_tensor.LabelTensor`,
            :class:`~pina.graph.Graph`, or :class:`~torch_geometric.data.Data`.
        """
        if cls != InputTargetCondition:
            return super().__new__(cls)

        if isinstance(input, (torch.Tensor, LabelTensor)) and isinstance(
            target, (torch.Tensor, LabelTensor)
        ):
            subclass = TensorInputTensorTargetCondition
            return subclass.__new__(subclass, input, target)
        if isinstance(input, (torch.Tensor, LabelTensor)) and isinstance(
            target, (Graph, Data, list, tuple)
        ):
            cls._check_graph_list_consistency(target)
            subclass = TensorInputGraphTargetCondition
            return subclass.__new__(subclass, input, target)

        if isinstance(input, (Graph, Data, list, tuple)) and isinstance(
            target, (torch.Tensor, LabelTensor)
        ):
            cls._check_graph_list_consistency(input)
            subclass = GraphInputTensorTargetCondition
            return subclass.__new__(subclass, input, target)

        if isinstance(input, (Graph, Data, list, tuple)) and isinstance(
            target, (Graph, Data, list, tuple)
        ):
            cls._check_graph_list_consistency(input)
            cls._check_graph_list_consistency(target)
            subclass = GraphInputGraphTargetCondition
            return subclass.__new__(subclass, input, target)

        raise ValueError(
            "Invalid input/target types. "
            "Please provide either torch_geometric.data.Data, Graph, "
            "LabelTensor or torch.Tensor objects."
        )

    def __init__(self, input, target):
        """
        Initialize the object by storing the ``input`` and ``target`` data.

        :param input: Input data for the condition.
        :type input: torch.Tensor | LabelTensor | Graph | Data | list[Graph] |
            list[Data] | tuple[Graph] | tuple[Data]
        :param target: Target data for the condition.
        :type target: torch.Tensor | LabelTensor | Graph | Data | list[Graph] |
            list[Data] | tuple[Graph] | tuple[Data]

        .. note::
            If either input or target consists of a list of
            :class:~pina.graph.Graph or :class:~torch_geometric.data.Data
            objects, all elements must have the same structure (matching
            keys and data types).
        """

        super().__init__()
        self._check_input_target_len(input, target)
        self.input = input
        self.target = target

    @staticmethod
    def _check_input_target_len(input, target):
        if isinstance(input, (Graph, Data)) or isinstance(
            target, (Graph, Data)
        ):
            return
        if len(input) != len(target):
            raise ValueError(
                "The input and target lists must have the same length."
            )


class TensorInputTensorTargetCondition(InputTargetCondition):
    """
    InputTargetCondition subclass for :class:`torch.Tensor` or
    :class:`~pina.label_tensor.LabelTensor` ``input`` and ``target`` data.
    """


class TensorInputGraphTargetCondition(InputTargetCondition):
    """
    InputTargetCondition subclass for :class:`torch.Tensor` or
    :class:`~pina.label_tensor.LabelTensor` ``input`` and
    :class:`~pina.graph.Graph` or :class:`~torch_geometric.data.Data` `target`
    data.
    """


class GraphInputTensorTargetCondition(InputTargetCondition):
    """
    InputTargetCondition subclass for :class:`~pina.graph.Graph` o
    :class:`~torch_geometric.data.Data` ``input`` and :class:`torch.Tensor` or
    :class:`~pina.label_tensor.LabelTensor` ``target`` data.
    """


class GraphInputGraphTargetCondition(InputTargetCondition):
    """
    InputTargetCondition subclass for :class:`~pina.graph.Graph`/
    :class:`~torch_geometric.data.Data` ``input`` and ``target`` data.
    """



"""Module for the Physics-Informed Neural Network Interface."""

from abc import ABCMeta, abstractmethod
import torch

from ..supervised_solver import SupervisedSolverInterface
from ...condition import (
    InputTargetCondition,
    InputEquationCondition,
    DomainEquationCondition,
)


class PINNInterface(SupervisedSolverInterface, metaclass=ABCMeta):
    """
    Base class for Physics-Informed Neural Network (PINN) solvers, implementing
    the :class:`~pina.solver.solver.SolverInterface` class.

    The `PINNInterface` class can be used to define PINNs that work with one or
    multiple optimizers and/or models. By default, it is compatible with
    problems defined by :class:`~pina.problem.abstract_problem.AbstractProblem`,
    and users can choose the problem type the solver is meant to address.
    """

    accepted_conditions_types = (
        InputTargetCondition,
        InputEquationCondition,
        DomainEquationCondition,
    )

    def __init__(self, **kwargs):
        """
        Initialization of the :class:`PINNInterface` class.

        :param AbstractProblem problem: The problem to be solved.
        :param torch.nn.Module loss: The loss function to be minimized.
            If ``None``, the :class:`torch.nn.MSELoss` loss is used.
            Default is `None`.
        :param kwargs: Additional keyword arguments to be passed to the
            :class:`~pina.solver.supervised_solver.SupervisedSolverInterface`
            class.
        """
        kwargs["use_lt"] = True
        super().__init__(**kwargs)

        # current condition name
        self.__metric = None

    def optimization_cycle(self, batch, loss_residuals=None):
        """
        The optimization cycle for the PINN solver.

        This method allows to call `_run_optimization_cycle` with the physics
        loss as argument, thus distinguishing the training step from the
        validation and test steps.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :return: The losses computed for all conditions in the batch, casted
            to a subclass of :class:`torch.Tensor`. It should return a dict
            containing the condition name and the associated scalar loss.
        :rtype: dict
        """
        # which losses to use
        if loss_residuals is None:
            loss_residuals = self.loss_phys
        # compute optimization cycle
        condition_loss = {}
        for condition_name, points in batch:
            self.__metric = condition_name
            # if equations are passed
            if "target" not in points:
                input_pts = points["input"]
                condition = self.problem.conditions[condition_name]
                loss = loss_residuals(
                    input_pts.requires_grad_(), condition.equation
                )
            # if data are passed
            else:
                input_pts = points["input"]
                output_pts = points["target"]
                loss = self.loss_data(
                    input=input_pts.requires_grad_(), target=output_pts
                )
            # append loss
            condition_loss[condition_name] = loss
        return condition_loss

    @torch.set_grad_enabled(True)
    def validation_step(self, batch):
        """
        The validation step for the PINN solver. It returns the average residual
        computed with the ``loss`` function not aggregated.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :return: The loss of the validation step.
        :rtype: torch.Tensor
        """
        return super().validation_step(
            batch, loss_residuals=self._residual_loss
        )

    @torch.set_grad_enabled(True)
    def test_step(self, batch):
        """
        The test step for the PINN solver. It returns the average residual
        computed with the ``loss`` function not aggregated.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :return: The loss of the test step.
        :rtype: torch.Tensor
        """
        return super().test_step(batch, loss_residuals=self._residual_loss)

    def loss_data(self, input, target):
        """
        Compute the data loss for the PINN solver by evaluating the loss
        between the network's output and the true solution. This method should
        be overridden by the derived class.

        :param LabelTensor input: The input to the neural network.
        :param LabelTensor target: The target to compare with the
            network's output.
        :return: The supervised loss, averaged over the number of observations.
        :rtype: LabelTensor
        :raises NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError(
            "PINN is being used in a supervised learning context, but the "
            "'loss_data' method has not been implemented. "
        )

    @abstractmethod
    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the physics-informed solver based on the
        provided samples and equation. This method must be overridden in
        subclasses. It distinguishes different types of PINN solvers.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation.
        :return: The computed physics loss.
        :rtype: LabelTensor
        """

    def compute_residual(self, samples, equation):
        """
        Compute the residuals of the equation.

        :param LabelTensor samples: The samples to evaluate the loss.
        :param EquationInterface equation: The governing equation.
        :return: The residual of the solution of the model.
        :rtype: LabelTensor
        """
        try:
            residual = equation.residual(samples, self.forward(samples))
        except TypeError:
            # this occurs when the function has three inputs (inverse problem)
            residual = equation.residual(
                samples, self.forward(samples), self._params
            )
        return residual

    def _residual_loss(self, samples, equation):
        """
        Computes the physics loss for the physics-informed solver based on the
        provided samples and equation. This method should never be overridden
        by the user, if not intentionally,
        since it is used internally to compute validation loss.


        :param LabelTensor samples: The samples to evaluate the loss.
        :param EquationInterface equation: The governing equation.
        :return: The residual loss.
        :rtype: torch.Tensor
        """
        residuals = self.compute_residual(samples, equation)
        return self._loss_fn(residuals, torch.zeros_like(residuals))

    @property
    def current_condition_name(self):
        """
        The current condition name.

        :return: The current condition name.
        :rtype: str
        """
        return self.__metric



import torch
import pytest

from pina.adaptive_function import (
    AdaptiveReLU,
    AdaptiveSigmoid,
    AdaptiveTanh,
    AdaptiveSiLU,
    AdaptiveMish,
    AdaptiveELU,
    AdaptiveCELU,
    AdaptiveGELU,
    AdaptiveSoftmin,
    AdaptiveSoftmax,
    AdaptiveSIREN,
    AdaptiveExp,
)


adaptive_function = (
    AdaptiveReLU,
    AdaptiveSigmoid,
    AdaptiveTanh,
    AdaptiveSiLU,
    AdaptiveMish,
    AdaptiveELU,
    AdaptiveCELU,
    AdaptiveGELU,
    AdaptiveSoftmin,
    AdaptiveSoftmax,
    AdaptiveSIREN,
    AdaptiveExp,
)
x = torch.rand(10, requires_grad=True)


@pytest.mark.parametrize("Func", adaptive_function)
def test_constructor(Func):
    if Func.__name__ == "AdaptiveExp":
        # simple
        Func()
        # setting values
        af = Func(alpha=1.0, beta=2.0)
        assert af.alpha.requires_grad
        assert af.beta.requires_grad
        assert af.alpha == 1.0
        assert af.beta == 2.0
    else:
        # simple
        Func()
        # setting values
        af = Func(alpha=1.0, beta=2.0, gamma=3.0)
        assert af.alpha.requires_grad
        assert af.beta.requires_grad
        assert af.gamma.requires_grad
        assert af.alpha == 1.0
        assert af.beta == 2.0
        assert af.gamma == 3.0

    # fixed variables
    af = Func(alpha=1.0, beta=2.0, fixed=["alpha"])
    assert af.alpha.requires_grad is False
    assert af.beta.requires_grad
    assert af.alpha == 1.0
    assert af.beta == 2.0

    with pytest.raises(TypeError):
        Func(alpha=1.0, beta=2.0, fixed=["delta"])

    with pytest.raises(ValueError):
        Func(alpha="s")
        Func(alpha=1)


@pytest.mark.parametrize("Func", adaptive_function)
def test_forward(Func):
    af = Func()
    af(x)


@pytest.mark.parametrize("Func", adaptive_function)
def test_backward(Func):
    af = Func()
    y = af(x)
    y.mean().backward()



"""Module for the R3Refinement callback."""

import torch
from torch import nn
from torch.nn.modules.loss import _Loss
from .refinement_interface import RefinementInterface
from ...label_tensor import LabelTensor
from ...utils import check_consistency
from ...loss import LossInterface


class R3Refinement(RefinementInterface):
    """
    PINA Implementation of an R3 Refinement Callback.
    """

    def __init__(
        self, sample_every, residual_loss=nn.L1Loss, condition_to_update=None
    ):
        """
        This callback implements the R3 (Retain-Resample-Release) routine for
        sampling new points based on adaptive search.
        The algorithm incrementally accumulates collocation points in regions
        of high PDE residuals, and releases those with low residuals.
        Points are sampled uniformly in all regions where sampling is needed.

        .. seealso::

            Original Reference: Daw, Arka, et al. *Mitigating Propagation
            Failures in Physics-informed Neural Networks
            using Retain-Resample-Release (R3) Sampling. (2023)*.
            DOI: `10.48550/arXiv.2207.02338
            <https://doi.org/10.48550/arXiv.2207.02338>`_

        :param int sample_every: Frequency for sampling.
        :param loss: Loss function
        :type loss: LossInterface | ~torch.nn.modules.loss._Loss
        :param condition_to_update: The conditions to update during the
            refinement process. If None, all conditions with a conditions will
            be updated. Default is None.
        :type condition_to_update: list(str) | tuple(str) | str
        :raises ValueError: If the condition_to_update is not a string or
            iterable of strings.
        :raises TypeError: If the residual_loss is not a subclass of
            torch.nn.Module.


        Example:
            >>> r3_callback = R3Refinement(sample_every=5)
        """
        super().__init__(sample_every, condition_to_update)
        # check consistency loss
        check_consistency(residual_loss, (LossInterface, _Loss), subclass=True)
        self.loss_fn = residual_loss(reduction="none")

    def sample(self, current_points, condition_name, solver):
        """
        Sample new points based on the R3 refinement strategy.

        :param current_points: Current points in the domain.
        :param condition_name: Name of the condition to update.
        :param PINNInterface solver: The solver object.
        :return: New points sampled based on the R3 strategy.
        :rtype: LabelTensor
        """
        # Compute residuals for the given condition (average over fields)
        condition = solver.problem.conditions[condition_name]
        target = solver.compute_residual(
            current_points.requires_grad_(True), condition.equation
        )
        residuals = self.loss_fn(target, torch.zeros_like(target)).mean(
            dim=tuple(range(1, target.ndim))
        )

        # Prepare new points
        labels = current_points.labels
        domain_name = solver.problem.conditions[condition_name].domain
        domain = solver.problem.domains[domain_name]
        num_old_points = self.initial_population_size[condition_name]
        mask = (residuals > residuals.mean()).flatten()

        if mask.any():  # Use high-residual points
            pts = current_points[mask]
            pts.labels = labels
            retain_pts = len(pts)
            samples = domain.sample(num_old_points - retain_pts, "random")
            return LabelTensor.cat([pts, samples])
        return domain.sample(num_old_points, "random")



"""Module for enforcing type hints in Python functions."""

import inspect
import typing
import logging


def enforce_types(func):
    """
    Function decorator to enforce type hints at runtime.

    This decorator checks the types of the arguments and of the return value of
    the decorated function against the type hints specified in the function
    signature. If the types do not match, a TypeError is raised.
    Type checking is only performed when the logging level is set to `DEBUG`.

    :param Callable func: The function to be decorated.
    :return: The decorated function with enforced type hints.
    :rtype: Callable

    :Example:

    >>> @enforce_types
        def dummy_function(a: int, b: float) -> float:
        ... return a+b

        # This always works.
        dummy_function(1, 2.0)

        # This raises a TypeError for the second argument, if logging is set to
        # `DEBUG`.
        dummy_function(1, "Hello, world!")


    >>> @enforce_types
        def dummy_function2(a: int, right: bool) -> float:
        ... if right:
        ...     return float(a)
        ... else:
        ...     return "Hello, world!"

        # This always works.
        dummy_function2(1, right=True)

        # This raises a TypeError for the return value if logging is set to
        # `DEBUG`.
        dummy_function2(1, right=False)
    """

    def wrapper(*args, **kwargs):
        """
        Wrapper function to enforce type hints.

        :param tuple args: Positional arguments passed to the function.
        :param dict kwargs: Keyword arguments passed to the function.
        :raises TypeError: If the argument or return type does not match the
            specified type hints.
        :return: The result of the decorated function.
        :rtype: Any
        """
        level = logging.getLevelName(logging.getLogger().getEffectiveLevel())

        # Enforce type hints only in debug mode
        if level != "DEBUG":
            return func(*args, **kwargs)

        # Get the type hints for the function arguments
        hints = typing.get_type_hints(func)
        sig = inspect.signature(func)
        bound = sig.bind(*args, **kwargs)
        bound.apply_defaults()

        for arg_name, arg_value in bound.arguments.items():
            expected_type = hints.get(arg_name)
            if expected_type and not isinstance(arg_value, expected_type):
                raise TypeError(
                    f"Argument '{arg_name}' must be {expected_type.__name__}, "
                    f"but got {type(arg_value).__name__}!"
                )

        # Get the type hints for the return values
        return_type = hints.get("return")
        result = func(*args, **kwargs)

        if return_type and not isinstance(result, return_type):
            raise TypeError(
                f"Return value must be {return_type.__name__}, "
                f"but got {type(result).__name__}!"
            )

        return result

    return wrapper



"""Module for the DeepONet and MIONet model classes."""

from functools import partial
import torch
from torch import nn
from ..utils import check_consistency, is_function


class MIONet(torch.nn.Module):
    """
    MIONet model class.

    The MIONet is a general architecture for learning operators, which map
    functions to functions. It can be trained with both Supervised and
    Physics-Informed learning strategies.

    .. seealso::

        **Original reference**: Jin, P., Meng, S., and Lu L. (2022).
        *MIONet: Learning multiple-input operators via tensor product.*
        SIAM Journal on Scientific Computing 44.6 (2022): A3490-A351
        DOI: `10.1137/22M1477751 <https://doi.org/10.1137/22M1477751>`_
    """

    def __init__(
        self,
        networks,
        aggregator="*",
        reduction="+",
        scale=True,
        translation=True,
    ):
        """
        Initialization of the :class:`MIONet` class.

        :param dict networks: The neural networks to use as models. The ``dict``
            takes as key a neural network, and as value the list of indeces to
            extract from the input variable in the forward pass of the neural
            network. If a ``list[int]`` is passed, the corresponding columns of
            the inner most entries are extracted. If a ``list[str]`` is passed
            the variables of the corresponding
            :class:`~pina.label_tensor.LabelTensor` are extracted.
            Each :class:`torch.nn.Module` model has to take as input either a
            :class:`~pina.label_tensor.LabelTensor` or a :class:`torch.Tensor`.
            Default implementation consists of several branch nets and one
            trunk nets.
        :param aggregator: The aggregator to be used to aggregate component-wise
            partial results from the modules in ``networks``. Available
            aggregators include: sum: ``+``, product: ``*``, mean: ``mean``,
            min: ``min``, max: ``max``. Default is ``*``.
        :type aggregator: str or Callable
        :param reduction: The reduction to be used to reduce the aggregated
            result of the modules in ``networks`` to the desired output
            dimension. Available reductions include: sum: ``+``, product: ``*``,
            mean: ``mean``, min: ``min``, max: ``max``. Default is ``+``.
        :type reduction: str or Callable
        :param bool scale: If ``True``, the final output is scaled before being
            returned in the forward pass. Default is ``True``.
        :param bool translation: If ``True``, the final output is translated
            before being returned in the forward pass.  Default is ``True``.
        :raises ValueError: If the passed networks have not the same output
            dimension.

        .. warning::
            No checks are performed in the forward pass to verify if the input
            is instance of either :class:`~pina.label_tensor.LabelTensor` or
            :class:`torch.Tensor`. In general, in case of a
            :class:`~pina.label_tensor.LabelTensor`, both a ``list[int]`` or a
            ``list[str]`` can be passed as ``networks`` dict values.
            Differently, in case of a :class:`torch.Tensor`, only a
            ``list[int]`` can be passed as ``networks`` dict values.

        :Example:
            >>> branch_net1 = FeedForward(input_dimensons=1,
            ... output_dimensions=10)
            >>> branch_net2 = FeedForward(input_dimensons=2,
            ... output_dimensions=10)
            >>> trunk_net = FeedForward(input_dimensons=1, output_dimensions=10)
            >>> networks = {branch_net1 : ['x'],
                            branch_net2 : ['x', 'y'],
            ...             trunk_net : ['z']}
            >>> model = MIONet(networks=networks,
            ...                reduction='+',
            ...                aggregator='*')
            >>> model
            MIONet(
            (models): ModuleList(
                (0): FeedForward(
                (model): Sequential(
                    (0): Linear(in_features=1, out_features=20, bias=True)
                    (1): Tanh()
                    (2): Linear(in_features=20, out_features=20, bias=True)
                    (3): Tanh()
                    (4): Linear(in_features=20, out_features=10, bias=True)
                )
                )
                (1): FeedForward(
                (model): Sequential(
                    (0): Linear(in_features=2, out_features=20, bias=True)
                    (1): Tanh()
                    (2): Linear(in_features=20, out_features=20, bias=True)
                    (3): Tanh()
                    (4): Linear(in_features=20, out_features=10, bias=True)
                )
                )
                (2): FeedForward(
                (model): Sequential(
                    (0): Linear(in_features=1, out_features=20, bias=True)
                    (1): Tanh()
                    (2): Linear(in_features=20, out_features=20, bias=True)
                    (3): Tanh()
                    (4): Linear(in_features=20, out_features=10, bias=True)
                )
                )
            )
            )
        """
        super().__init__()

        # check type consistency
        check_consistency(networks, dict)
        check_consistency(scale, bool)
        check_consistency(translation, bool)

        # check trunk branch nets consistency
        shapes = []
        for key, value in networks.items():
            check_consistency(value, (str, int))
            check_consistency(key, torch.nn.Module)
            input_ = torch.rand(10, len(value))
            shapes.append(key(input_).shape[-1])

        if not all(map(lambda x: x == shapes[0], shapes)):
            raise ValueError(
                "The passed networks have not the same output dimension."
            )

        # assign trunk and branch net with their input indeces
        self.models = torch.nn.ModuleList(networks.keys())
        self._indeces = networks.values()

        # initializie aggregation
        self._init_aggregator(aggregator=aggregator)
        self._init_reduction(reduction=reduction)

        # scale and translation
        self._scale = (
            torch.nn.Parameter(torch.tensor([1.0]))
            if scale
            else torch.tensor([1.0])
        )
        self._trasl = (
            torch.nn.Parameter(torch.tensor([1.0]))
            if translation
            else torch.tensor([1.0])
        )

    @staticmethod
    def _symbol_functions(**kwargs):
        """
        Return a dictionary of functions that can be used as aggregators or
        reductions.

        :param dict kwargs: Additional parameters.
        :return: A dictionary of functions.
        :rtype: dict
        """
        return {
            "+": partial(torch.sum, **kwargs),
            "*": partial(torch.prod, **kwargs),
            "mean": partial(torch.mean, **kwargs),
            "min": lambda x: torch.min(x, **kwargs).values,
            "max": lambda x: torch.max(x, **kwargs).values,
        }

    def _init_aggregator(self, aggregator):
        """
        Initialize the aggregator.

        :param aggregator: The aggregator to be used to aggregate.
        :type aggregator: str or Callable
        :raises ValueError: If the aggregator is not supported.
        """
        aggregator_funcs = self._symbol_functions(dim=2)
        if aggregator in aggregator_funcs:
            aggregator_func = aggregator_funcs[aggregator]
        elif isinstance(aggregator, nn.Module) or is_function(aggregator):
            aggregator_func = aggregator
        else:
            raise ValueError(f"Unsupported aggregation: {str(aggregator)}")

        self._aggregator = aggregator_func
        self._aggregator_type = aggregator

    def _init_reduction(self, reduction):
        """
        Initialize the reduction.

        :param reduction: The reduction to be used.
        :type reduction: str or Callable
        :raises ValueError: If the reduction is not supported.
        """
        reduction_funcs = self._symbol_functions(dim=-1)
        if reduction in reduction_funcs:
            reduction_func = reduction_funcs[reduction]
        elif isinstance(reduction, nn.Module) or is_function(reduction):
            reduction_func = reduction
        else:
            raise ValueError(f"Unsupported reduction: {reduction}")

        self._reduction = reduction_func
        self._reduction_type = reduction

    def _get_vars(self, x, indeces):
        """
        Extract the variables from the input tensor.

        :param x: The input tensor.
        :type x: LabelTensor | torch.Tensor
        :param indeces: The indeces to extract.
        :type indeces: list[int] | list[str]
        :raises RuntimeError: If failing to extract the variables.
        :raises RuntimeError: If failing to extract the right indeces.
        :return: The extracted variables.
        :rtype: LabelTensor | torch.Tensor
        """
        if isinstance(indeces[0], str):
            try:
                return x.extract(indeces)
            except AttributeError as e:
                raise RuntimeError(
                    "Not possible to extract input variables from tensor."
                    " Ensure that the passed tensor is a LabelTensor or"
                    " pass list of integers to extract variables. For"
                    " more information refer to warning in the documentation."
                ) from e
        elif isinstance(indeces[0], int):
            return x[..., indeces]
        else:
            raise RuntimeError(
                "Not able to extract right indeces for tensor."
                " For more information refer to warning in the documentation."
            )

    def forward(self, x):
        """
        Forward pass for the :class:`MIONet` model.

        :param x: The input tensor.
        :type x: LabelTensor | torch.Tensor
        :return: The output tensor.
        :rtype: LabelTensor | torch.Tensor
        """

        # forward pass
        output_ = [
            model(self._get_vars(x, indeces))
            for model, indeces in zip(self.models, self._indeces)
        ]

        # aggregation
        aggregated = self._aggregator(torch.dstack(output_))

        # reduce
        output_ = self._reduction(aggregated)
        if self._reduction_type in self._symbol_functions(dim=-1):
            output_ = output_.reshape(-1, 1)

        # scale and translate
        output_ *= self._scale
        output_ += self._trasl

        return output_

    @property
    def aggregator(self):
        """
        The aggregator function.

        :return: The aggregator function.
        :rtype: str or Callable
        """
        return self._aggregator

    @property
    def reduction(self):
        """
        The reduction function.

        :return: The reduction function.
        :rtype: str or Callable
        """
        return self._reduction

    @property
    def scale(self):
        """
        The scale factor.

        :return: The scale factor.
        :rtype: torch.Tensor
        """
        return self._scale

    @property
    def translation(self):
        """
        The translation factor.

        :return: The translation factor.
        :rtype: torch.Tensor
        """
        return self._trasl

    @property
    def indeces_variables_extracted(self):
        """
        The input indeces for each model in form of list.

        :return: The indeces for each model.
        :rtype: list
        """
        return self._indeces

    @property
    def model(self):
        """
        The models in form of list.

        :return: The models.
        :rtype: list[torch.nn.Module]
        """
        return self._indeces


class DeepONet(MIONet):
    """
    DeepONet model class.

    The MIONet is a general architecture for learning operators, which map
    functions to functions. It can be trained with both Supervised and
    Physics-Informed learning strategies.

    .. seealso::

        **Original reference**: Lu, L., Jin, P., Pang, G. et al.
        *Learning nonlinear operators via DeepONet based on the universal
        approximation theorem of operator*.
        Nat Mach Intell 3, 218-229 (2021).
        DOI: `10.1038/s42256-021-00302-5
        <https://doi.org/10.1038/s42256-021-00302-5>`_

    """

    def __init__(
        self,
        branch_net,
        trunk_net,
        input_indeces_branch_net,
        input_indeces_trunk_net,
        aggregator="*",
        reduction="+",
        scale=True,
        translation=True,
    ):
        """
        Initialization of the :class:`DeepONet` class.

        :param torch.nn.Module branch_net: The neural network to use as branch
            model. It has to take as input either a
            :class:`~pina.label_tensor.LabelTensor` or a :class:`torch.Tensor`.
            The output dimension has to be the same as that of ``trunk_net``.
        :param torch.nn.Module trunk_net: The neural network to use as trunk
            model. It has to take as input either a
            :class:`~pina.label_tensor.LabelTensor` or a :class:`torch.Tensor`.
            The output dimension has to be the same as that of ``branch_net``.
        :param input_indeces_branch_net: List of indeces to extract from the
            input variable of the ``branch_net``.
            If a list of ``int`` is passed, the corresponding columns of the
            inner most entries are extracted. If a list of ``str`` is passed the
            variables of the corresponding
            :class:`~pina.label_tensor.LabelTensor` are extracted.
        :type input_indeces_branch_net: list[int] | list[str]
        :param input_indeces_trunk_net: List of indeces to extract from the
            input variable of the ``trunk_net``.
            If a list of ``int`` is passed, the corresponding columns of the
            inner most entries are extracted. If a list of ``str`` is passed the
            variables of the corresponding
            :class:`~pina.label_tensor.LabelTensor` are extracted.
        :type input_indeces_trunk_net: list[int] | list[str]
        :param aggregator: The aggregator to be used to aggregate component-wise
            partial results from the modules in ``networks``. Available
            aggregators include: sum: ``+``, product: ``*``, mean: ``mean``,
            min: ``min``, max: ``max``. Default is ``*``.
        :type aggregator: str or Callable
        :param reduction: The reduction to be used to reduce the aggregated
            result of the modules in ``networks`` to the desired output
            dimension. Available reductions include: sum: ``+``, product: ``*``,
            mean: ``mean``, min: ``min``, max: ``max``. Default is ``+``.
        :type reduction: str or Callable
        :param bool scale: If ``True``, the final output is scaled before being
            returned in the forward pass. Default is ``True``.
        :param bool translation: If ``True``, the final output is translated
            before being returned in the forward pass.  Default is ``True``.

        .. warning::
            In the forward pass we do not check if the input is instance of
            :py:obj:`pina.label_tensor.LabelTensor` or :class:`torch.Tensor`.
            A general rule is that for a :py:obj:`pina.label_tensor.LabelTensor`
            input both list of integers and list of strings can be passed for
            ``input_indeces_branch_net`` and ``input_indeces_trunk_net``.
            Differently, for a :class:`torch.Tensor` only a list of integers can
            be passed for ``input_indeces_branch_net`` and
            ``input_indeces_trunk_net``.

        .. warning::
            No checks are performed in the forward pass to verify if the input
            is instance of either :class:`~pina.label_tensor.LabelTensor` or
            :class:`torch.Tensor`. In general, in case of a
            :class:`~pina.label_tensor.LabelTensor`, both a ``list[int]`` or a
            ``list[str]`` can be passed as ``input_indeces_branch_net`` and
            ``input_indeces_trunk_net``. Differently, in case of a
            :class:`torch.Tensor`, only a ``list[int]`` can be passed.

        :Example:
            >>> branch_net = FeedForward(input_dimensons=1,
            ... output_dimensions=10)
            >>> trunk_net = FeedForward(input_dimensons=1, output_dimensions=10)
            >>> model = DeepONet(branch_net=branch_net,
            ...                  trunk_net=trunk_net,
            ...                  input_indeces_branch_net=['x'],
            ...                  input_indeces_trunk_net=['t'],
            ...                  reduction='+',
            ...                  aggregator='*')
            >>> model
            DeepONet(
            (trunk_net): FeedForward(
                (model): Sequential(
                (0): Linear(in_features=1, out_features=20, bias=True)
                (1): Tanh()
                (2): Linear(in_features=20, out_features=20, bias=True)
                (3): Tanh()
                (4): Linear(in_features=20, out_features=10, bias=True)
                )
            )
            (branch_net): FeedForward(
                (model): Sequential(
                (0): Linear(in_features=1, out_features=20, bias=True)
                (1): Tanh()
                (2): Linear(in_features=20, out_features=20, bias=True)
                (3): Tanh()
                (4): Linear(in_features=20, out_features=10, bias=True)
                )
            )
            )
        """
        networks = {
            branch_net: input_indeces_branch_net,
            trunk_net: input_indeces_trunk_net,
        }
        super().__init__(
            networks=networks,
            aggregator=aggregator,
            reduction=reduction,
            scale=scale,
            translation=translation,
        )

    def forward(self, x):
        """
        Forward pass for the :class:`DeepONet` model.

        :param x: The input tensor.
        :type x: LabelTensor | torch.Tensor
        :return: The output tensor.
        :rtype: LabelTensor | torch.Tensor
        """
        return super().forward(x)

    @property
    def branch_net(self):
        """
        The branch net of the DeepONet.

        :return: The branch net.
        :rtype: torch.nn.Module
        """
        return self.models[0]

    @property
    def trunk_net(self):
        """
        The trunk net of the DeepONet.

        :return: The trunk net.
        :rtype: torch.nn.Module
        """
        return self.models[1]



"""Module for the Spline model class."""

import torch
from ..utils import check_consistency


class Spline(torch.nn.Module):
    """
    Spline model class.
    """

    def __init__(self, order=4, knots=None, control_points=None) -> None:
        """
        Initialization of the :class:`Spline` class.

        :param int order: The order of the spline. Default is ``4``.
        :param torch.Tensor knots: The tensor representing knots. If ``None``,
            the knots will be initialized automatically. Default is ``None``.
        :param torch.Tensor control_points: The control points. Default is
            ``None``.
        :raises ValueError: If the order is negative.
        :raises ValueError: If both knots and control points are ``None``.
        :raises ValueError: If the knot tensor is not one-dimensional.
        """
        super().__init__()

        check_consistency(order, int)

        if order < 0:
            raise ValueError("Spline order cannot be negative.")
        if knots is None and control_points is None:
            raise ValueError("Knots and control points cannot be both None.")

        self.order = order
        self.k = order - 1

        if knots is not None and control_points is not None:
            self.knots = knots
            self.control_points = control_points

        elif knots is not None:
            print("Warning: control points will be initialized automatically.")
            print("         experimental feature")

            self.knots = knots
            n = len(knots) - order
            self.control_points = torch.nn.Parameter(
                torch.zeros(n), requires_grad=True
            )

        elif control_points is not None:
            print("Warning: knots will be initialized automatically.")
            print("         experimental feature")

            self.control_points = control_points

            n = len(self.control_points) - 1
            self.knots = {
                "type": "auto",
                "min": 0,
                "max": 1,
                "n": n + 2 + self.order,
            }

        else:
            raise ValueError("Knots and control points cannot be both None.")

        if self.knots.ndim != 1:
            raise ValueError("Knot vector must be one-dimensional.")

    def basis(self, x, k, i, t):
        """
        Recursive method to compute the basis functions of the spline.

        :param torch.Tensor x: The points to be evaluated.
        :param int k: The spline degree.
        :param int i: The index of the interval.
        :param torch.Tensor t: The tensor of knots.
        :return: The basis functions evaluated at x
        :rtype: torch.Tensor
        """

        if k == 0:
            a = torch.where(
                torch.logical_and(t[i] <= x, x < t[i + 1]), 1.0, 0.0
            )
            if i == len(t) - self.order - 1:
                a = torch.where(x == t[-1], 1.0, a)
            a.requires_grad_(True)
            return a

        if t[i + k] == t[i]:
            c1 = torch.tensor([0.0] * len(x), requires_grad=True)
        else:
            c1 = (x - t[i]) / (t[i + k] - t[i]) * self.basis(x, k - 1, i, t)

        if t[i + k + 1] == t[i + 1]:
            c2 = torch.tensor([0.0] * len(x), requires_grad=True)
        else:
            c2 = (
                (t[i + k + 1] - x)
                / (t[i + k + 1] - t[i + 1])
                * self.basis(x, k - 1, i + 1, t)
            )

        return c1 + c2

    @property
    def control_points(self):
        """
        The control points of the spline.

        :return: The control points.
        :rtype: torch.Tensor
        """
        return self._control_points

    @control_points.setter
    def control_points(self, value):
        """
        Set the control points of the spline.

        :param value: The control points.
        :type value: torch.Tensor | dict
        :raises ValueError: If invalid value is passed.
        """
        if isinstance(value, dict):
            if "n" not in value:
                raise ValueError("Invalid value for control_points")
            n = value["n"]
            dim = value.get("dim", 1)
            value = torch.zeros(n, dim)

        if not isinstance(value, torch.Tensor):
            raise ValueError("Invalid value for control_points")
        self._control_points = torch.nn.Parameter(value, requires_grad=True)

    @property
    def knots(self):
        """
        The knots of the spline.

        :return: The knots.
        :rtype: torch.Tensor
        """
        return self._knots

    @knots.setter
    def knots(self, value):
        """
        Set the knots of the spline.

        :param value: The knots.
        :type value: torch.Tensor | dict
        :raises ValueError: If invalid value is passed.
        """
        if isinstance(value, dict):

            type_ = value.get("type", "auto")
            min_ = value.get("min", 0)
            max_ = value.get("max", 1)
            n = value.get("n", 10)

            if type_ == "uniform":
                value = torch.linspace(min_, max_, n + self.k + 1)
            elif type_ == "auto":
                initial_knots = torch.ones(self.order + 1) * min_
                final_knots = torch.ones(self.order + 1) * max_

                if n < self.order + 1:
                    value = torch.concatenate((initial_knots, final_knots))
                elif n - 2 * self.order + 1 == 1:
                    value = torch.Tensor([(max_ + min_) / 2])
                else:
                    value = torch.linspace(min_, max_, n - 2 * self.order - 1)

                value = torch.concatenate((initial_knots, value, final_knots))

        if not isinstance(value, torch.Tensor):
            raise ValueError("Invalid value for knots")

        self._knots = value

    def forward(self, x):
        """
        Forward pass for the :class:`Spline` model.

        :param torch.Tensor x: The input tensor.
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        t = self.knots
        k = self.k
        c = self.control_points

        basis = map(lambda i: self.basis(x, k, i, t)[:, None], range(len(c)))
        y = (torch.cat(list(basis), dim=1) * c).sum(axis=1)

        return y



"""Module for the Low Rank Neural Operator Block class."""

import torch

from ...utils import check_consistency


class LowRankBlock(torch.nn.Module):
    """
    The inner block of the Low Rank Neural Operator.

    .. seealso::

        **Original reference**: Kovachki, N., Li, Z., Liu, B.,
        Azizzadenesheli, K., Bhattacharya, K., Stuart, A., & Anandkumar, A.
        (2023). *Neural operator: Learning maps between function
        spaces with applications to PDEs*. Journal of Machine Learning
        Research, 24(89), 1-97.
    """

    def __init__(
        self,
        input_dimensions,
        embedding_dimenion,
        rank,
        inner_size=20,
        n_layers=2,
        func=torch.nn.Tanh,
        bias=True,
    ):
        r"""
        Initialization of the :class:`LowRankBlock` class.

        :param int input_dimensions: The input dimension of the field.
        :param int embedding_dimenion: The embedding dimension of the field.
        :param int rank: The rank of the low rank approximation. The expected
            value is :math:`2d`, where :math:`d` is the rank of each basis
            function.
        :param int inner_size: The number of neurons for each hidden layer in
            the basis function neural network. Default is ``20``.
        :param int n_layers: The number of hidden layers in the basis function
            neural network. Default is ``2``.
        :param func: The activation function. If a list is passed, it must have
            the same length as ``n_layers``. If a single function is passed, it
            is used for all layers, except for the last one.
            Default is :class:`torch.nn.Tanh`.
        :type func: torch.nn.Module | list[torch.nn.Module]
        :param bool bias: If ``True`` bias is considered for the basis function
            neural network. Default is ``True``.
        """
        super().__init__()
        from ..feed_forward import FeedForward

        # Assignment (check consistency inside FeedForward)
        self._basis = FeedForward(
            input_dimensions=input_dimensions,
            output_dimensions=2 * rank * embedding_dimenion,
            inner_size=inner_size,
            n_layers=n_layers,
            func=func,
            bias=bias,
        )
        self._nn = torch.nn.Linear(embedding_dimenion, embedding_dimenion)

        check_consistency(rank, int)
        self._rank = rank
        self._func = func()

    def forward(self, x, coords):
        r"""
        Forward pass of the block. It performs an affine transformation of the
        field, followed by a low rank approximation. The latter is performed by
        means of a dot product of the basis :math:`\psi^{(i)}` with the vector
        field :math:`v` to compute coefficients used to expand
        :math:`\phi^{(i)}`, evaluated in the spatial input :math:`x`.

        :param torch.Tensor x: The input tensor for performing the computation.
        :param torch.Tensor coords: The coordinates for which the field is
            evaluated to perform the computation.
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        # extract basis
        coords = coords.as_subclass(torch.Tensor)
        basis = self._basis(coords)
        # reshape [B, N, D, 2*rank]
        shape = list(basis.shape[:-1]) + [-1, 2 * self.rank]
        basis = basis.reshape(shape)
        # divide
        psi = basis[..., : self.rank]
        phi = basis[..., self.rank :]
        # compute dot product
        coeff = torch.einsum("...dr,...d->...r", psi, x)
        # expand the basis
        expansion = torch.einsum("...r,...dr->...d", coeff, phi)
        # apply linear layer and return
        return self._func(self._nn(x) + expansion)

    @property
    def rank(self):
        """
        The basis rank.

        :return: The basis rank.
        :rtype: int
        """
        return self._rank



"""Module for the Equation Interface."""

from abc import ABCMeta, abstractmethod


class EquationInterface(metaclass=ABCMeta):
    """
    Abstract base class for equations.

    Equations in PINA simplify the training process. When defining a problem,
    each equation passed to a :class:`~pina.condition.condition.Condition`
    object must be either an :class:`~pina.equation.equation.Equation` or a
    :class:`~pina.equation.system_equation.SystemEquation` instance.

    An :class:`~pina.equation.equation.Equation` is a wrapper for a callable
    function, while :class:`~pina.equation.system_equation.SystemEquation`
    wraps a list of callable functions. To streamline code writing, PINA
    provides a diverse set of pre-implemented equations, such as
    :class:`~pina.equation.equation_factory.FixedValue`,
    :class:`~pina.equation.equation_factory.FixedGradient`, and many others.
    """

    @abstractmethod
    def residual(self, input_, output_, params_):
        """
        Abstract method to compute the residual of an equation.

        :param LabelTensor input_: Input points where the equation is evaluated.
        :param LabelTensor output_: Output tensor, eventually produced by a
            :class:`torch.nn.Module` instance.
        :param dict params_: Dictionary of unknown parameters, associated with a
            :class:`~pina.problem.inverse_problem.InverseProblem` instance.
        :return: The computed residual of the equation.
        :rtype: LabelTensor
        """



"""Formulation of the inverse Poisson problem in a square domain."""

import requests
import torch
from io import BytesIO
from ... import Condition
from ... import LabelTensor
from ...operator import laplacian
from ...domain import CartesianDomain
from ...equation import Equation, FixedValue
from ...problem import SpatialProblem, InverseProblem


def laplace_equation(input_, output_, params_):
    """
    Implementation of the laplace equation.

    :param LabelTensor input_: Input data of the problem.
    :param LabelTensor output_: Output data of the problem.
    :param dict params_: Parameters of the problem.
    :return: The residual of the laplace equation.
    :rtype: LabelTensor
    """
    force_term = torch.exp(
        -2 * (input_.extract(["x"]) - params_["mu1"]) ** 2
        - 2 * (input_.extract(["y"]) - params_["mu2"]) ** 2
    )
    delta_u = laplacian(output_, input_, components=["u"], d=["x", "y"])
    return delta_u - force_term


# URL of the file
url = "https://github.com/mathLab/PINA/raw/refs/heads/master/tutorials/tutorial7/data/pts_0.5_0.5"
# Download the file
response = requests.get(url)
response.raise_for_status()
file_like_object = BytesIO(response.content)
# Set the data
input_data = LabelTensor(
    torch.load(file_like_object, weights_only=False).tensor.detach(),
    ["x", "y", "mu1", "mu2"],
)

# URL of the file
url = "https://github.com/mathLab/PINA/raw/refs/heads/master/tutorials/tutorial7/data/pinn_solution_0.5_0.5"
# Download the file
response = requests.get(url)
response.raise_for_status()
file_like_object = BytesIO(response.content)
# Set the data
output_data = LabelTensor(
    torch.load(file_like_object, weights_only=False).tensor.detach(), ["u"]
)


class InversePoisson2DSquareProblem(SpatialProblem, InverseProblem):
    r"""
    Implementation of the inverse 2-dimensional Poisson problem in the square
    domain :math:`[0, 1] \times [0, 1]`,
    with unknown parameter domain :math:`[-1, 1] \times [-1, 1]`.

    :Example:
        >>> problem = InversePoisson2DSquareProblem()
    """

    output_variables = ["u"]
    x_min, x_max = -2, 2
    y_min, y_max = -2, 2
    spatial_domain = CartesianDomain({"x": [x_min, x_max], "y": [y_min, y_max]})
    unknown_parameter_domain = CartesianDomain({"mu1": [-1, 1], "mu2": [-1, 1]})

    domains = {
        "g1": CartesianDomain({"x": [x_min, x_max], "y": y_max}),
        "g2": CartesianDomain({"x": [x_min, x_max], "y": y_min}),
        "g3": CartesianDomain({"x": x_max, "y": [y_min, y_max]}),
        "g4": CartesianDomain({"x": x_min, "y": [y_min, y_max]}),
        "D": CartesianDomain({"x": [x_min, x_max], "y": [y_min, y_max]}),
    }

    conditions = {
        "g1": Condition(domain="g1", equation=FixedValue(0.0)),
        "g2": Condition(domain="g2", equation=FixedValue(0.0)),
        "g3": Condition(domain="g3", equation=FixedValue(0.0)),
        "g4": Condition(domain="g4", equation=FixedValue(0.0)),
        "D": Condition(domain="D", equation=Equation(laplace_equation)),
        "data": Condition(input=input_data, target=output_data),
    }



"""Module for the Stride class."""

import torch


class Stride:
    """
    Stride class for continous convolution.
    """

    def __init__(self, dict_):
        """
        Initialization of the :class:`Stride` class.

        :param dict dict_: Dictionary having as keys the domain size ``domain``,
            the starting position of the filter ``start``, the jump size for the
            filter ``jump``, and the direction of the filter ``direction``.
        """

        self._dict_stride = dict_
        self._stride_continuous = None
        self._stride_discrete = self._create_stride_discrete(dict_)

    def _create_stride_discrete(self, my_dict):
        """
        Create a tensor of positions where to apply the filter.

        :param dict my_dict_: Dictionary having as keys the domain size
            ``domain``, the starting position of the filter ``start``, the jump
            size for the filter ``jump``, and the direction of the filter
            ``direction``.
        :raises IndexError: Values in the dict must have all same length.
        :raises ValueError: Domain values must be greater than 0.
        :raises ValueError: Direction must be either equal to ``1``, ``-1`` or
            ``0``.
        :raises IndexError: Direction and jumps must be zero in the same index.
        :return: The positions for the filter
        :rtype: torch.Tensor

        :Example:

            >>> stride_dict = {
            ...     "domain": [4, 4],
            ...     "start": [-4, 2],
            ...     "jump": [2, 2],
            ...     "direction": [1, 1],
            ... }
            >>> Stride(stride_dict)
        """
        # we must check boundaries of the input as well
        domain, start, jumps, direction = my_dict.values()

        # checking
        if not all(len(s) == len(domain) for s in my_dict.values()):
            raise IndexError("Values in the dict must have all same length")

        if not all(v >= 0 for v in domain):
            raise ValueError("Domain values must be greater than 0")

        if not all(v in (0, -1, 1) for v in direction):
            raise ValueError("Direction must be either equal to 1, -1 or 0")

        seq_jumps = [i for i, e in enumerate(jumps) if e == 0]
        seq_direction = [i for i, e in enumerate(direction) if e == 0]

        if seq_direction != seq_jumps:
            raise IndexError(
                "Direction and jumps must have zero in the same index"
            )

        if seq_jumps:
            for i in seq_jumps:
                jumps[i] = domain[i]
                direction[i] = 1

        # creating the stride grid
        values_mesh = [
            torch.arange(0, i, step).float() for i, step in zip(domain, jumps)
        ]

        values_mesh = [
            single * dim for single, dim in zip(values_mesh, direction)
        ]

        mesh = torch.meshgrid(values_mesh)
        coordinates_mesh = [x.reshape(-1, 1) for x in mesh]

        stride = torch.cat(coordinates_mesh, dim=1) + torch.tensor(start)

        return stride



"""Old layers module, deprecated in 0.2.0."""

import warnings

from ..block import *
from ...utils import custom_warning_format

# back-compatibility 0.1
# Set the custom format for warnings
warnings.formatwarning = custom_warning_format
warnings.filterwarnings("always", category=DeprecationWarning)
warnings.warn(
    "'pina.model.layers' is deprecated and will be removed "
    "in future versions. Please use 'pina.model.block' instead.",
    DeprecationWarning,
)



import pytest
import logging
import math
from pina.type_checker import enforce_types


# Definition of a test function for arguments
@enforce_types
def foo_function1(a: int, b: float) -> float:
    return a + b


# Definition of a test function for return values
@enforce_types
def foo_function2(a: int, right: bool) -> float:
    if right:
        return float(a)
    else:
        return "Hello, world!"


def test_argument_type_checking():

    # Setting logging level to INFO, which should not trigger type checking
    logging.getLogger().setLevel(logging.INFO)

    # Both should work, even if the arguments are not of the expected type
    assert math.isclose(foo_function1(a=1, b=2.0), 3.0)
    assert math.isclose(foo_function1(a=1, b=2), 3.0)

    # Setting logging level to DEBUG, which should trigger type checking
    logging.getLogger().setLevel(logging.DEBUG)

    # The second should fail, as the second argument is an int
    assert math.isclose(foo_function1(a=1, b=2.0), 3.0)
    with pytest.raises(TypeError):
        foo_function1(a=1, b=2)


def test_return_type_checking():

    # Setting logging level to INFO, which should not trigger type checking
    logging.getLogger().setLevel(logging.INFO)

    # Both should work, even if the return value is not of the expected type
    assert math.isclose(foo_function2(a=1, right=True), 1.0)
    assert foo_function2(a=1, right=False) == "Hello, world!"

    # Setting logging level to DEBUG, which should trigger type checking
    logging.getLogger().setLevel(logging.DEBUG)

    # The second should fail, as the return value is a string
    assert math.isclose(foo_function2(a=1, right=True), 1.0)
    with pytest.raises(TypeError):
        foo_function2(a=1, right=False)



"""Formulation of a Supervised Problem in PINA."""

from ..abstract_problem import AbstractProblem
from ... import Condition


class SupervisedProblem(AbstractProblem):
    """
    Definition of a supervised-learning problem.

    This class provides a simple way to define a supervised problem
    using a single condition of type
    :class:`~pina.condition.input_target_condition.InputTargetCondition`.

    :Example:
        >>> import torch
        >>> input_data = torch.rand((100, 10))
        >>> output_data = torch.rand((100, 10))
        >>> problem = SupervisedProblem(input_data, output_data)
    """

    conditions = {}
    output_variables = None
    input_variables = None

    def __init__(
        self, input_, output_, input_variables=None, output_variables=None
    ):
        """
        Initialization of the :class:`SupervisedProblem` class.

        :param input_: Input data of the problem.
        :type input_: torch.Tensor | LabelTensor | Graph | Data
        :param output_: Output data of the problem.
        :type output_: torch.Tensor | LabelTensor | Graph | Data
        """
        # Set input and output variables
        self.input_variables = input_variables
        self.output_variables = output_variables
        # Set the condition
        self.conditions["data"] = Condition(input=input_, target=output_)
        super().__init__()



"""Module for the Loss Interface."""

from abc import ABCMeta, abstractmethod
from torch.nn.modules.loss import _Loss
import torch


class LossInterface(_Loss, metaclass=ABCMeta):
    """
    Abstract base class for all losses. All classes defining a loss function
    should inherit from this interface.
    """

    def __init__(self, reduction="mean"):
        """
        Initialization of the :class:`LossInterface` class.

        :param str reduction: The reduction method for the loss.
            Available options: ``none``, ``mean``, ``sum``.
            If ``none``, no reduction is applied. If ``mean``, the sum of the
            loss values is divided by the number of values. If ``sum``, the loss
            values are summed. Default is ``mean``.
        """
        super().__init__(reduction=reduction, size_average=None, reduce=None)

    @abstractmethod
    def forward(self, input, target):
        """
        Forward method of the loss function.

        :param torch.Tensor input: Input tensor from real data.
        :param torch.Tensor target: Model tensor output.
        """

    def _reduction(self, loss):
        """
        Apply the reduction to the loss.

        :param torch.Tensor loss: The tensor containing the pointwise losses.
        :raises ValueError: If the reduction method is not valid.
        :return: Reduced loss.
        :rtype: torch.Tensor
        """
        if self.reduction == "none":
            ret = loss
        elif self.reduction == "mean":
            ret = torch.mean(loss, keepdim=True, dim=-1)
        elif self.reduction == "sum":
            ret = torch.sum(loss, keepdim=True, dim=-1)
        else:
            raise ValueError(self.reduction + " is not valid")
        return ret



"""Module for the Ensemble solver classes."""

__all__ = [
    "DeepEnsembleSolverInterface",
    "DeepEnsembleSupervisedSolver",
    "DeepEnsemblePINN",
]

from .ensemble_solver_interface import DeepEnsembleSolverInterface
from .ensemble_supervised import DeepEnsembleSupervisedSolver
from .ensemble_pinn import DeepEnsemblePINN



"""Module for the Operation Interface."""

from abc import ABCMeta, abstractmethod
from .domain_interface import DomainInterface
from ..utils import check_consistency


class OperationInterface(DomainInterface, metaclass=ABCMeta):
    """
    Abstract class for set operations defined on geometric domains.
    """

    def __init__(self, geometries):
        """
        Initialization of the :class:`OperationInterface` class.

        :param list[DomainInterface] geometries: A list of instances of the
            :class:`~pina.domain.domain_interface.DomainInterface` class on
            which the set operation is performed.
        """
        # check consistency geometries
        check_consistency(geometries, DomainInterface)

        # check we are passing always different
        # geometries with the same labels.
        self._check_dimensions(geometries)

        # assign geometries
        self._geometries = geometries

    @property
    def sample_modes(self):
        """
        List of available sampling modes.

        :return: List of available sampling modes.
        :rtype: list[str]
        """
        return ["random"]

    @property
    def geometries(self):
        """
        The domains on which to perform the set operation.

        :return: The domains on which to perform the set operation.
        :rtype: list[DomainInterface]
        """
        return self._geometries

    @property
    def variables(self):
        """
        List of variables of the domain.

        :return: List of variables of the domain.
        :rtype: list[str]
        """
        variables = []
        for geom in self.geometries:
            variables += geom.variables
        return sorted(list(set(variables)))

    @abstractmethod
    def is_inside(self, point, check_border=False):
        """
        Abstract method to check if a point lies inside the resulting domain
        after performing the set operation.

        :param LabelTensor point: Point to be checked.
        :param bool check_border: If ``True``, the border is considered inside
            the resulting domain. Default is ``False``.
        :return: ``True`` if the point is inside the domain,
            ``False`` otherwise.
        :rtype: bool
        """

    def _check_dimensions(self, geometries):
        """
        Check if the dimensions of the geometries are consistent.

        :param list[DomainInterface] geometries: Domains to be checked.
        :raises NotImplementedError: If the dimensions of the geometries are not
            consistent.
        """
        for geometry in geometries:
            if geometry.variables != geometries[0].variables:
                raise NotImplementedError(
                    "The geometries need to have same dimensions and labels."
                )



"""Module for the InputEquationCondition class and its subclasses."""

from torch_geometric.data import Data
from .condition_interface import ConditionInterface
from ..label_tensor import LabelTensor
from ..graph import Graph
from ..utils import check_consistency
from ..equation.equation_interface import EquationInterface


class InputEquationCondition(ConditionInterface):
    """
    Condition defined by input data and an equation. This condition can be
    used in a Physics Informed problems. Based on the type of the input,
    different condition implementations are available:

    - :class:`InputTensorEquationCondition`: For \
        :class:`~pina.label_tensor.LabelTensor` input data.
    - :class:`InputGraphEquationCondition`: For :class:`~pina.graph.Graph` \
        input data.
    """

    __slots__ = ["input", "equation"]
    _avail_input_cls = (LabelTensor, Graph, list, tuple)
    _avail_equation_cls = EquationInterface

    def __new__(cls, input, equation):
        """
        Instantiate the appropriate subclass of :class:`InputEquationCondition`
        based on the type of ``input``.

        :param input: Input data for the condition.
        :type input: LabelTensor | Graph |  list[Graph] | tuple[Graph]
        :param EquationInterface equation: Equation object containing the
            equation function.
        :return: Subclass of InputEquationCondition, based on the input type.
        :rtype: pina.condition.input_equation_condition.
            InputTensorEquationCondition |
            pina.condition.input_equation_condition.InputGraphEquationCondition

        :raises ValueError: If input is not of type
            :class:`~pina.label_tensor.LabelTensor`, :class:`~pina.graph.Graph`.
        """

        # If the class is already a subclass, return the instance
        if cls != InputEquationCondition:
            return super().__new__(cls)

        # Instanciate the correct subclass
        if isinstance(input, (Graph, Data, list, tuple)):
            subclass = InputGraphEquationCondition
            cls._check_graph_list_consistency(input)
            subclass._check_label_tensor(input)
            return subclass.__new__(subclass, input, equation)
        if isinstance(input, LabelTensor):
            subclass = InputTensorEquationCondition
            return subclass.__new__(subclass, input, equation)

        # If the input is not a LabelTensor or a Graph object raise an error
        raise ValueError(
            "The input data object must be a LabelTensor or a Graph object."
        )

    def __init__(self, input, equation):
        """
        Initialize the object by storing the input data and equation object.

        :param input: Input data for the condition.
        :type input: LabelTensor | Graph |
            list[Graph] | tuple[Graph]
        :param EquationInterface equation: Equation object containing the
            equation function.

        .. note::
            If ``input`` consists of a list of :class:`~pina.graph.Graph` or
            :class:`~torch_geometric.data.Data`, all elements must have the same
            structure (keys and data types)
        """

        super().__init__()
        self.input = input
        self.equation = equation

    def __setattr__(self, key, value):
        if key == "input":
            check_consistency(value, self._avail_input_cls)
            InputEquationCondition.__dict__[key].__set__(self, value)
        elif key == "equation":
            check_consistency(value, self._avail_equation_cls)
            InputEquationCondition.__dict__[key].__set__(self, value)
        elif key in ("_problem"):
            super().__setattr__(key, value)


class InputTensorEquationCondition(InputEquationCondition):
    """
    InputEquationCondition subclass for :class:`~pina.label_tensor.LabelTensor`
    input data.
    """


class InputGraphEquationCondition(InputEquationCondition):
    """
    InputEquationCondition subclass for :class:`~pina.graph.Graph` input data.
    """

    @staticmethod
    def _check_label_tensor(input):
        """
        Check if at least one :class:`~pina.label_tensor.LabelTensor` is present
        in the :class:`~pina.graph.Graph` object.

        :param input: Input data.
        :type input: torch.Tensor | Graph | Data

        :raises ValueError: If the input data object does not contain at least
            one LabelTensor.
        """

        # Store the fist element of the list/tuple if input is a list/tuple
        # it is anougth to check the first element because all elements must
        # have the same type and structure (already checked)
        data = input[0] if isinstance(input, (list, tuple)) else input

        # Check if the input data contains at least one LabelTensor
        for v in data.values():
            if isinstance(v, LabelTensor):
                return
        raise ValueError(
            "The input data object must contain at least one LabelTensor."
        )



import torch
import pytest
from pina import LabelTensor
from pina.operator import grad, div, laplacian, advection


class Function(object):

    def __iter__(self):
        functions = [
            (
                getattr(self, f"{name}_input"),
                getattr(self, f"{name}"),
                getattr(self, f"{name}_grad"),
                getattr(self, f"{name}_div"),
                getattr(self, f"{name}_lap"),
            )
            for name in [
                "scalar_scalar",
                "scalar_vector",
                "vector_scalar",
                "vector_vector",
            ]
        ]
        return iter(functions)

    # Scalar to scalar function
    @staticmethod
    def scalar_scalar(x):
        return x**2

    @staticmethod
    def scalar_scalar_grad(x):
        return 2 * x

    @staticmethod
    def scalar_scalar_div(x):
        return 2 * x

    @staticmethod
    def scalar_scalar_lap(x):
        return 2 * torch.ones_like(x)

    @staticmethod
    def scalar_scalar_input():
        input_ = torch.rand((20, 1), requires_grad=True)
        return LabelTensor(input_, ["x"])

    # Scalar to vector function
    @staticmethod
    def scalar_vector(x):
        u = x**2
        v = x**3 + x
        return torch.cat((u, v), dim=-1)

    @staticmethod
    def scalar_vector_grad(x):
        u = 2 * x
        v = 3 * x**2 + 1
        return torch.cat((u, v), dim=-1)

    @staticmethod
    def scalar_vector_div(x):
        return ValueError

    @staticmethod
    def scalar_vector_lap(x):
        u = 2 * torch.ones_like(x)
        v = 6 * x
        return torch.cat((u, v), dim=-1)

    @staticmethod
    def scalar_vector_input():
        input_ = torch.rand((20, 1), requires_grad=True)
        return LabelTensor(input_, ["x"])

    # Vector to scalar function
    @staticmethod
    def vector_scalar(x):
        return torch.prod(x**2, dim=-1, keepdim=True)

    @staticmethod
    def vector_scalar_grad(x):
        return 2 * torch.prod(x**2, dim=-1, keepdim=True) / x

    @staticmethod
    def vector_scalar_div(x):
        return ValueError

    @staticmethod
    def vector_scalar_lap(x):
        return 2 * torch.sum(
            torch.prod(x**2, dim=-1, keepdim=True) / x**2,
            dim=-1,
            keepdim=True,
        )

    @staticmethod
    def vector_scalar_input():
        input_ = torch.rand((20, 2), requires_grad=True)
        return LabelTensor(input_, ["x", "yy"])

    # Vector to vector function
    @staticmethod
    def vector_vector(x):
        u = torch.prod(x**2, dim=-1, keepdim=True)
        v = torch.sum(x**2, dim=-1, keepdim=True)
        return torch.cat((u, v), dim=-1)

    @staticmethod
    def vector_vector_grad(x):
        u = 2 * torch.prod(x**2, dim=-1, keepdim=True) / x
        v = 2 * x
        return torch.cat((u, v), dim=-1)

    @staticmethod
    def vector_vector_div(x):
        u = 2 * torch.prod(x**2, dim=-1, keepdim=True) / x[..., 0]
        v = 2 * x[..., 1]
        return u + v

    @staticmethod
    def vector_vector_lap(x):
        u = torch.sum(
            2 * torch.prod(x**2, dim=-1, keepdim=True) / x**2,
            dim=-1,
            keepdim=True,
        )
        v = 2 * x.shape[-1] * torch.ones_like(u)
        return torch.cat((u, v), dim=-1)

    @staticmethod
    def vector_vector_input():
        input_ = torch.rand((20, 2), requires_grad=True)
        return LabelTensor(input_, ["x", "yy"])


@pytest.mark.parametrize(
    "f",
    Function(),
    ids=["scalar_scalar", "scalar_vector", "vector_scalar", "vector_vector"],
)
def test_gradient(f):

    # Unpack the function
    func_input, func, func_grad, _, _ = f

    # Define input and output
    input_ = func_input()
    output_ = func(input_)
    labels = [f"u{i}" for i in range(output_.shape[-1])]
    output_ = LabelTensor(output_, labels)

    # Compute the true gradient and the pina gradient
    pina_grad = grad(output_=output_, input_=input_)
    true_grad = func_grad(input_)

    # Check the shape and labels of the gradient
    n_components = len(output_.labels) * len(input_.labels)
    assert pina_grad.shape == (*output_.shape[:-1], n_components)
    assert pina_grad.labels == [
        f"d{c}d{i}" for c in output_.labels for i in input_.labels
    ]

    # Compare the values
    assert torch.allclose(pina_grad, true_grad)

    # Test if labels are handled correctly
    grad(output_=output_, input_=input_, components=output_.labels[0])
    grad(output_=output_, input_=input_, d=input_.labels[0])

    # Should fail if input not a LabelTensor
    with pytest.raises(TypeError):
        grad(output_=output_, input_=input_.tensor)

    # Should fail if output not a LabelTensor
    with pytest.raises(TypeError):
        grad(output_=output_.tensor, input_=input_)

    # Should fail for non-existent input labels
    with pytest.raises(RuntimeError):
        grad(output_=output_, input_=input_, d=["x", "y"])

    # Should fail for non-existent output labels
    with pytest.raises(RuntimeError):
        grad(output_=output_, input_=input_, components=["a", "b", "c"])


@pytest.mark.parametrize(
    "f",
    Function(),
    ids=["scalar_scalar", "scalar_vector", "vector_scalar", "vector_vector"],
)
def test_divergence(f):

    # Unpack the function
    func_input, func, _, func_div, _ = f

    # Define input and output
    input_ = func_input()
    output_ = func(input_)
    labels = [f"u{i}" for i in range(output_.shape[-1])]
    output_ = LabelTensor(output_, labels)

    # Scalar to vector or vector to scalar functions
    if func_div(input_) == ValueError:
        with pytest.raises(ValueError):
            div(output_=output_, input_=input_)

    # Scalar to scalar or vector to vector functions
    else:
        # Compute the true divergence and the pina divergence
        pina_div = div(output_=output_, input_=input_)
        true_div = func_div(input_)

        # Check the shape and labels of the divergence
        assert pina_div.shape == (*output_.shape[:-1], 1)
        tmp_labels = [
            f"d{c}d{d_}" for c, d_ in zip(output_.labels, input_.labels)
        ]
        assert pina_div.labels == ["+".join(tmp_labels)]

        # Compare the values
        assert torch.allclose(pina_div, true_div)

        # Test if labels are handled correctly. Performed in a single call to
        # avoid components and d having different lengths.
        div(
            output_=output_,
            input_=input_,
            components=output_.labels[0],
            d=input_.labels[0],
        )

        # Should fail if input not a LabelTensor
        with pytest.raises(TypeError):
            div(output_=output_, input_=input_.tensor)

        # Should fail if output not a LabelTensor
        with pytest.raises(TypeError):
            div(output_=output_.tensor, input_=input_)

        # Should fail for non-existent labels
        with pytest.raises(RuntimeError):
            div(output_=output_, input_=input_, d=["x", "y"])

        with pytest.raises(RuntimeError):
            div(output_=output_, input_=input_, components=["a", "b", "c"])


@pytest.mark.parametrize(
    "f",
    Function(),
    ids=["scalar_scalar", "scalar_vector", "vector_scalar", "vector_vector"],
)
def test_laplacian(f):

    # Unpack the function
    func_input, func, _, _, func_lap = f

    # Define input and output
    input_ = func_input()
    output_ = func(input_)
    labels = [f"u{i}" for i in range(output_.shape[-1])]
    output_ = LabelTensor(output_, labels)

    # Compute the true laplacian and the pina laplacian
    pina_lap = laplacian(output_=output_, input_=input_)
    true_lap = func_lap(input_)

    # Check the shape and labels of the laplacian
    assert pina_lap.shape == output_.shape
    assert pina_lap.labels == [f"dd{l}" for l in output_.labels]

    # Compare the values
    assert torch.allclose(pina_lap, true_lap)

    # Test if labels are handled correctly
    laplacian(output_=output_, input_=input_, components=output_.labels[0])
    laplacian(output_=output_, input_=input_, d=input_.labels[0])

    # Should fail if input not a LabelTensor
    with pytest.raises(TypeError):
        laplacian(output_=output_, input_=input_.tensor)

    # Should fail if output not a LabelTensor
    with pytest.raises(TypeError):
        laplacian(output_=output_.tensor, input_=input_)

    # Should fail for non-existent input labels
    with pytest.raises(RuntimeError):
        laplacian(output_=output_, input_=input_, d=["x", "y"])

    # Should fail for non-existent output labels
    with pytest.raises(RuntimeError):
        laplacian(output_=output_, input_=input_, components=["a", "b", "c"])


def test_advection_scalar():

    # Define 3-dimensional input
    input_ = torch.rand((20, 3), requires_grad=True)
    input_ = LabelTensor(input_, ["x", "y", "z"])

    # Define 3-dimensional velocity field and quantity to be advected
    velocity = torch.rand((20, 3), requires_grad=True)
    field = torch.sum(input_**2, dim=-1, keepdim=True)

    # Combine velocity and field into a LabelTensor
    labels = ["ux", "uy", "uz", "c"]
    output_ = LabelTensor(torch.cat((velocity, field), dim=1), labels)

    # Compute the pina advection
    components = ["c"]
    pina_adv = advection(
        output_=output_,
        input_=input_,
        velocity_field=["ux", "uy", "uz"],
        components=components,
        d=["x", "y", "z"],
    )

    # Compute the true advection
    grads = 2 * input_
    true_adv = torch.sum(grads * velocity, dim=grads.ndim - 1, keepdim=True)

    # Check the shape, labels, and value of the advection
    assert pina_adv.shape == (*output_.shape[:-1], len(components))
    assert pina_adv.labels == ["adv_c"]
    assert torch.allclose(pina_adv, true_adv)

    # Should fail if input not a LabelTensor
    with pytest.raises(TypeError):
        advection(
            output_=output_,
            input_=input_.tensor,
            velocity_field=["ux", "uy", "uz"],
        )

    # Should fail if output not a LabelTensor
    with pytest.raises(TypeError):
        advection(
            output_=output_.tensor,
            input_=input_,
            velocity_field=["ux", "uy", "uz"],
        )

    # Should fail for non-existent input labels
    with pytest.raises(RuntimeError):
        advection(
            output_=output_,
            input_=input_,
            d=["x", "a"],
            velocity_field=["ux", "uy", "uz"],
        )

    # Should fail for non-existent output labels
    with pytest.raises(RuntimeError):
        advection(
            output_=output_,
            input_=input_,
            components=["a", "b", "c"],
            velocity_field=["ux", "uy", "uz"],
        )

    # Should fail if velocity_field labels are not present in the output labels
    with pytest.raises(RuntimeError):
        advection(
            output_=output_,
            input_=input_,
            velocity_field=["ux", "uy", "nonexistent"],
            components=["c"],
        )

    # Should fail if velocity_field dimensionality does not match input tensor
    with pytest.raises(RuntimeError):
        advection(
            output_=output_,
            input_=input_,
            velocity_field=["ux", "uy"],
            components=["c"],
        )


def test_advection_vector():

    # Define 3-dimensional input
    input_ = torch.rand((20, 3), requires_grad=True)
    input_ = LabelTensor(input_, ["x", "y", "z"])

    # Define 3-dimensional velocity field
    velocity = torch.rand((20, 3), requires_grad=True)

    # Define 2-dimensional field to be advected
    field_1 = torch.sum(input_**2, dim=-1, keepdim=True)
    field_2 = torch.sum(input_**3, dim=-1, keepdim=True)

    # Combine velocity and field into a LabelTensor
    labels = ["ux", "uy", "uz", "c1", "c2"]
    output_ = LabelTensor(
        torch.cat((velocity, field_1, field_2), dim=1), labels
    )

    # Compute the pina advection
    components = ["c1", "c2"]
    pina_adv = advection(
        output_=output_,
        input_=input_,
        velocity_field=["ux", "uy", "uz"],
        components=components,
        d=["x", "y", "z"],
    )

    # Compute the true gradients of the fields "c1", "c2"
    grads1 = 2 * input_
    grads2 = 3 * input_**2

    # Compute the true advection for each field
    true_adv1 = torch.sum(grads1 * velocity, dim=grads1.ndim - 1, keepdim=True)
    true_adv2 = torch.sum(grads2 * velocity, dim=grads2.ndim - 1, keepdim=True)
    true_adv = torch.cat((true_adv1, true_adv2), dim=-1)

    # Check the shape, labels, and value of the advection
    assert pina_adv.shape == (*output_.shape[:-1], len(components))
    assert pina_adv.labels == ["adv_c1", "adv_c2"]
    assert torch.allclose(pina_adv, true_adv)

    # Should fail if input not a LabelTensor
    with pytest.raises(TypeError):
        advection(
            output_=output_,
            input_=input_.tensor,
            velocity_field=["ux", "uy", "uz"],
        )

    # Should fail if output not a LabelTensor
    with pytest.raises(TypeError):
        advection(
            output_=output_.tensor,
            input_=input_,
            velocity_field=["ux", "uy", "uz"],
        )

    # Should fail for non-existent input labels
    with pytest.raises(RuntimeError):
        advection(
            output_=output_,
            input_=input_,
            d=["x", "a"],
            velocity_field=["ux", "uy", "uz"],
        )

    # Should fail for non-existent output labels
    with pytest.raises(RuntimeError):
        advection(
            output_=output_,
            input_=input_,
            components=["a", "b", "c"],
            velocity_field=["ux", "uy", "uz"],
        )

    # Should fail if velocity_field labels are not present in the output labels
    with pytest.raises(RuntimeError):
        advection(
            output_=output_,
            input_=input_,
            velocity_field=["ux", "uy", "nonexistent"],
            components=["c"],
        )

    # Should fail if velocity_field dimensionality does not match input tensor
    with pytest.raises(RuntimeError):
        advection(
            output_=output_,
            input_=input_,
            velocity_field=["ux", "uy"],
            components=["c"],
        )



import torch
import pytest
from pina.problem.zoo import Poisson2DSquareProblem as Poisson
from pina import LabelTensor
from pina.domain import Union
from pina.domain import CartesianDomain
from pina.condition import (
    Condition,
    InputTargetCondition,
    DomainEquationCondition,
)


def test_discretise_domain():
    n = 10
    poisson_problem = Poisson()
    boundaries = ["g1", "g2", "g3", "g4"]
    poisson_problem.discretise_domain(n, "grid", domains=boundaries)
    for b in boundaries:
        assert poisson_problem.discretised_domains[b].shape[0] == n
    poisson_problem.discretise_domain(n, "random", domains=boundaries)
    for b in boundaries:
        assert poisson_problem.discretised_domains[b].shape[0] == n

    poisson_problem.discretise_domain(n, "grid", domains=["D"])
    assert poisson_problem.discretised_domains["D"].shape[0] == n**2
    poisson_problem.discretise_domain(n, "random", domains=["D"])
    assert poisson_problem.discretised_domains["D"].shape[0] == n

    poisson_problem.discretise_domain(n, "latin", domains=["D"])
    assert poisson_problem.discretised_domains["D"].shape[0] == n

    poisson_problem.discretise_domain(n, "lh", domains=["D"])
    assert poisson_problem.discretised_domains["D"].shape[0] == n

    poisson_problem.discretise_domain(n)


def test_variables_correct_order_sampling():
    n = 10
    poisson_problem = Poisson()
    poisson_problem.discretise_domain(n, "grid", domains=["D"])
    assert poisson_problem.discretised_domains["D"].labels == sorted(
        poisson_problem.input_variables
    )

    poisson_problem.discretise_domain(n, "grid", domains=["D"])
    assert poisson_problem.discretised_domains["D"].labels == sorted(
        poisson_problem.input_variables
    )


def test_input_pts():
    n = 10
    poisson_problem = Poisson()
    poisson_problem.discretise_domain(n, "grid")
    assert sorted(list(poisson_problem.input_pts.keys())) == sorted(
        list(poisson_problem.conditions.keys())
    )


def test_collected_data():
    n = 10
    poisson_problem = Poisson()
    poisson_problem.discretise_domain(n, "grid")
    assert sorted(list(poisson_problem.collected_data.keys())) == sorted(
        list(poisson_problem.conditions.keys())
    )


def test_add_points():
    poisson_problem = Poisson()
    poisson_problem.discretise_domain(0, "random", domains=["D"])
    new_pts = LabelTensor(torch.tensor([[0.5, -0.5]]), labels=["x", "y"])
    poisson_problem.add_points({"D": new_pts})
    assert torch.isclose(
        poisson_problem.discretised_domains["D"].extract("x"),
        new_pts.extract("x"),
    )
    assert torch.isclose(
        poisson_problem.discretised_domains["D"].extract("y"),
        new_pts.extract("y"),
    )


@pytest.mark.parametrize("mode", ["random", "grid"])
def test_custom_sampling_logic(mode):
    poisson_problem = Poisson()
    sampling_rules = {
        "x": {"n": 100, "mode": mode},
        "y": {"n": 50, "mode": mode},
    }
    poisson_problem.discretise_domain(sample_rules=sampling_rules)
    for domain in ["g1", "g2", "g3", "g4"]:
        assert poisson_problem.discretised_domains[domain].shape[0] == 100 * 50
        assert poisson_problem.discretised_domains[domain].labels == ["x", "y"]


@pytest.mark.parametrize("mode", ["random", "grid"])
def test_wrong_custom_sampling_logic(mode):
    d2 = CartesianDomain({"x": [1, 2], "y": [0, 1]})
    poisson_problem = Poisson()
    poisson_problem.domains["D"] = Union([poisson_problem.domains["D"], d2])
    sampling_rules = {
        "x": {"n": 100, "mode": mode},
        "y": {"n": 50, "mode": mode},
    }
    with pytest.raises(RuntimeError):
        poisson_problem.discretise_domain(sample_rules=sampling_rules)


def test_aggregate_data():
    poisson_problem = Poisson()
    poisson_problem.conditions["data"] = Condition(
        input=LabelTensor(torch.tensor([[0.0, 1.0]]), labels=["x", "y"]),
        target=LabelTensor(torch.tensor([[0.0]]), labels=["u"]),
    )
    poisson_problem.discretise_domain(0, "random", domains="all")
    poisson_problem.collect_data()
    assert isinstance(poisson_problem.collected_data, dict)
    for name, conditions in poisson_problem.conditions.items():
        assert name in poisson_problem.collected_data.keys()
        if isinstance(conditions, InputTargetCondition):
            assert "input" in poisson_problem.collected_data[name].keys()
            assert "target" in poisson_problem.collected_data[name].keys()
        elif isinstance(conditions, DomainEquationCondition):
            assert "input" in poisson_problem.collected_data[name].keys()
            assert "target" not in poisson_problem.collected_data[name].keys()
            assert "equation" in poisson_problem.collected_data[name].keys()



"""Module for the Orthogonal Block class."""

import torch
from ...utils import check_consistency


class OrthogonalBlock(torch.nn.Module):
    """
    Orthogonal Block.

    This block transforms an input tensor of shape :math:`[N, M]` into a tensor
    of the same shape whose columns are orthonormal. The block performs the
    Gram Schmidt orthogonalization, see
    `here <https://en.wikipedia.org/wiki/Gram%E2%80%93Schmidt_process>` for
    details.
    """

    def __init__(self, dim=-1, requires_grad=True):
        """
        Initialization of the :class:`OrthogonalBlock` class.

        :param int dim: The dimension on which orthogonalization is performed.
            If ``-1``, the orthogonalization is performed on the last dimension.
            Default is ``-1``.
        :param bool requires_grad: If ``True``, the gradients are computed
            during the backward pass. Default is ``True``
        """
        super().__init__()
        # store dim
        self.dim = dim
        # store requires_grad
        check_consistency(requires_grad, bool)
        self._requires_grad = requires_grad

    def forward(self, X):
        """
        Forward pass.

        :param torch.Tensor X: The input tensor to orthogonalize.
        :raises Warning: If the chosen dimension is greater than the other
            dimensions in the input.
        :return: The orthonormal tensor.
        :rtype: torch.Tensor
        """
        # check dim is less than all the other dimensions
        if X.shape[self.dim] > min(X.shape):
            raise Warning(
                "The dimension where to orthogonalize is greater"
                " than the other dimensions"
            )

        result = torch.zeros_like(X, requires_grad=self._requires_grad)
        X_0 = torch.select(X, self.dim, 0).clone()
        result_0 = X_0 / torch.linalg.norm(X_0)
        result = self._differentiable_copy(result, 0, result_0)

        # iterate over the rest of the basis with Gram-Schmidt
        for i in range(1, X.shape[self.dim]):
            v = torch.select(X, self.dim, i).clone()
            for j in range(i):
                vj = torch.select(result, self.dim, j).clone()
                v = v - torch.sum(v * vj, dim=self.dim, keepdim=True) * vj
            # result_i = torch.select(result, self.dim, i)
            result_i = v / torch.linalg.norm(v)
            result = self._differentiable_copy(result, i, result_i)
        return result

    def _differentiable_copy(self, result, idx, value):
        """
        Perform a differentiable copy operation.

        :param torch.Tensor result: The tensor where values are be copied to.
        :param int idx: The index along the specified dimension where the
            values are copied.
        :param torch.Tensor value: The tensor value to copy into ``result``.
        :return: A new tensor with the copied values.
        :rtype: torch.Tensor
        """
        return result.index_copy(
            self.dim, torch.tensor([idx]), value.unsqueeze(self.dim)
        )

    @property
    def dim(self):
        """
        The dimension along which operations are performed.

        :return: The current dimension value.
        :rtype: int
        """
        return self._dim

    @dim.setter
    def dim(self, value):
        """
        Set the dimension along which operations are performed.

        :param value: The dimension to be set. Must be either ``0``, ``1``, or
            ``-1``.
        :type value: int
        :raises IndexError: If the provided dimension is not ``0``, ``1``, or
            ``-1``.
        """
        # check consistency
        check_consistency(value, int)
        if value not in [0, 1, -1]:
            raise IndexError(
                "Dimension out of range (expected to be in "
                f"range of [-1, 1], but got {value})"
            )
        # assign value
        self._dim = value

    @property
    def requires_grad(self):
        """
        Indicates whether gradient computation is required for operations
        on the tensors.

        :return: ``True`` if gradients are required, ``False`` otherwise.
        :rtype: bool
        """
        return self._requires_grad



"""Module for the System of Equation."""

import torch
from .equation_interface import EquationInterface
from .equation import Equation
from ..utils import check_consistency


class SystemEquation(EquationInterface):
    """
    Implementation of the System of Equations. Every ``equation`` passed to a
    :class:`~pina.condition.condition.Condition` object must be either a
    :class:`~pina.equation.equation.Equation` or a
    :class:`~pina.equation.system_equation.SystemEquation` instance.
    """

    def __init__(self, list_equation, reduction=None):
        """
        Initialization of the :class:`SystemEquation` class.

        :param Callable equation: A ``torch`` callable function used to compute
            the residual of a mathematical equation.
        :param str reduction: The reduction method to aggregate the residuals of
            each equation. Available options are: ``None``, ``mean``, ``sum``,
            ``callable``.
            If ``None``, no reduction is applied. If ``mean``, the output sum is
            divided by the number of elements in the output. If ``sum``, the
            output is summed. ``callable`` is a user-defined callable function
            to perform reduction, no checks guaranteed. Default is ``None``.
        :raises NotImplementedError: If the reduction is not implemented.
        """
        check_consistency([list_equation], list)

        # equations definition
        self.equations = []
        for _, equation in enumerate(list_equation):
            self.equations.append(Equation(equation))

        # possible reduction
        if reduction == "mean":
            self.reduction = torch.mean
        elif reduction == "sum":
            self.reduction = torch.sum
        elif (reduction is None) or callable(reduction):
            self.reduction = reduction
        else:
            raise NotImplementedError(
                "Only mean and sum reductions implemented."
            )

    def residual(self, input_, output_, params_=None):
        """
        Compute the residual for each equation in the system of equations and
        aggregate it according to the ``reduction`` specified in the
        ``__init__`` method.

        :param LabelTensor input_: Input points where each equation of the
            system is evaluated.
        :param LabelTensor output_: Output tensor, eventually produced by a
            :class:`torch.nn.Module` instance.
        :param dict params_: Dictionary of unknown parameters, associated with a
            :class:`~pina.problem.inverse_problem.InverseProblem` instance.
            If the equation is not related to a
            :class:`~pina.problem.inverse_problem.InverseProblem` instance, the
            parameters must be initialized to ``None``. Default is ``None``.

        :return: The aggregated residuals of the system of equations.
        :rtype: LabelTensor
        """
        residual = torch.hstack(
            [
                equation.residual(input_, output_, params_)
                for equation in self.equations
            ]
        )

        if self.reduction is None:
            return residual

        return self.reduction(residual, dim=-1)



"""Module for the Domain Interface."""

from abc import ABCMeta, abstractmethod


class DomainInterface(metaclass=ABCMeta):
    """
    Abstract base class for geometric domains. All specific domain types should
    inherit from this class.
    """

    available_sampling_modes = ["random", "grid", "lh", "chebyshev", "latin"]

    @property
    @abstractmethod
    def sample_modes(self):
        """
        Abstract method defining sampling methods.
        """

    @property
    @abstractmethod
    def variables(self):
        """
        Abstract method returning the domain variables.
        """

    @sample_modes.setter
    def sample_modes(self, values):
        """
        Setter for the sample_modes property.

        :param values: Sampling modes to be set.
        :type values: str | list[str]
        :raises TypeError: Invalid sampling mode.
        """
        if not isinstance(values, (list, tuple)):
            values = [values]
        for value in values:
            if value not in DomainInterface.available_sampling_modes:
                raise TypeError(
                    f"mode {value} not valid. Expected at least "
                    "one in "
                    f"{DomainInterface.available_sampling_modes}."
                )

    @abstractmethod
    def sample(self):
        """
        Abstract method for the sampling routine.
        """

    @abstractmethod
    def is_inside(self, point, check_border=False):
        """
        Abstract method for checking if a point is inside the domain.

        :param LabelTensor point: Point to be checked.
        :param bool check_border: If ``True``, the border is considered inside
            the domain. Default is ``False``.
        """
